;;;; uuid-lib.scm
;;;; Kon Lovett, May '06
;;;; BSD

;;; Wrapper for Theodore Y. Ts'o OSF DCE 1.1 UUID Suite
;;; http://lib.sourceforge.net/
;;; Part of MacOS X & Linux (every modern unix?)

;; Issues
;;
;; - `uuid' is type, argument, struct tag variable, ...
;;
;; - variant means version or method

(declare
  (bound-to-procedure
    ##sys#signal-hook))

(module uuid-lib

(;export
  ;Common API
  make-uuid
  uuid?
  uuid-null?
  uuid-compare
  uuid=? uuid<? uuid>? uuid<=? uuid>=?
  uuid-hash
  uuid-copy
  uuid-clear!
  string->uuid uuid->string
  ;Specialized API
  check-uuid error-uuid ;FIXME add to uuid-ossp
  uuid-print ;FIXME add to uuid-ossp
  uuid-variant ;FIXME add to uuid-ossp
  uuid-generate
  uuid-parse
  uuid-unparse)

(import scheme)
(import (chicken base))
(import (chicken type))
(import (chicken fixnum))
(import (chicken format))
(import (chicken bitwise))
(import (chicken condition))
(import (chicken foreign))
(import (chicken memory))
(import (chicken gc))
(import (chicken locative))

;NOTE works w/o recourse to `define-record' for tag, so ok for tag "defining" module
(cond-expand
  ((or chicken-5.0 chicken-5.1)
    (define (set-record-printer! tag proc)
      (##sys#register-record-printer tag proc) ) )
  (else) )

(define-type uuid (struct uuid))
(define-type uuid-variant (or false symbol))
(define-type uuid-case (or false symbol))

(define-type error-argument-name (or string symbol))

(: uuid? (* -> boolean : uuid))
(: error-uuid (symbol * #!optional error-argument-name -> uuid))
(: check-uuid (symbol * #!optional error-argument-name -> uuid))

(: uuid-variant (uuid --> uuid-variant))

(: uuid-null? (* --> boolean)) ;not a `predicate'

(: uuid-hash (uuid #!optional fixnum fixnum --> fixnum))

(: uuid-compare (uuid uuid --> fixnum))
(: uuid=? (uuid uuid --> boolean))
(: uuid<? (uuid uuid --> boolean))
(: uuid>? (uuid uuid --> boolean))
(: uuid<=? (uuid uuid --> boolean))
(: uuid>=? (uuid uuid --> boolean))

(: uuid-clear! (uuid -> uuid))

(: uuid-copy (uuid --> uuid))

(: uuid-generate (#!optional uuid-variant --> uuid))
(: make-uuid (#!optional uuid-variant --> uuid))
(: uuid-parse ((or string symbol) --> uuid))

(: uuid-unparse (uuid #!optional uuid-case --> string))

(: string->uuid (string --> uuid))
(: uuid->string (uuid --> string))

;

(define-inline (fxabs n) (if (fx< n 0) (fxneg n) n))

(define-inline (fxnegative? n) (fx< n 0))
(define-inline (fxzero? n) (fx= n 0))
(define-inline (fxpositive? n) (fx> n 0))

;srfi-69

(foreign-declare "#define C_rnd_fix() (C_fix(rand()))")

(define-constant hash-default-bound 536870912)
(define hash-default-randomization (##core#inline "C_rnd_fix"))

;; Force Hash to Bounded Fixnum:

;@hsh : unsigned-integer
(define-inline (srfi69-hash-limit hsh bnd rnd)
  ;(assert (not (negative? hsh)))
  ;; use 32-bit mask to have identical hashes on 64-bit platforms
  (fxxor (modulo (bitwise-and hsh (foreign-value "C_MOST_POSITIVE_32_BIT_FIXNUM" int))
                 bnd)
         rnd) )

;;

#>
#include <uuid/uuid.h>
<#

(define UUID-SIZEOF (foreign-value "sizeof(uuid_t)" int))

(define-constant UUID-STRING-LENGTH 36)

(define uuid_is_null
  (foreign-lambda bool uuid_is_null nonnull-c-pointer))

(define uuid_compare
  (foreign-lambda int uuid_compare nonnull-c-pointer nonnull-c-pointer))

(define uuid_clear
  (foreign-lambda void uuid_clear nonnull-c-pointer))

(define uuid_copy
  (foreign-lambda void uuid_copy nonnull-c-pointer nonnull-c-pointer))

(define uuid_generate
  (foreign-lambda void uuid_generate nonnull-c-pointer))

(define uuid_generate_random
  (foreign-lambda void uuid_generate_random nonnull-c-pointer))

(define uuid_generate_time
  (foreign-lambda void uuid_generate_time nonnull-c-pointer))

(define uuid_parse
  (foreign-lambda int uuid_parse c-string nonnull-c-pointer))

(define uuid_unparse
  (foreign-lambda void uuid_unparse nonnull-c-pointer nonnull-c-pointer))

(define uuid_unparse_upper
  (foreign-lambda void uuid_unparse_upper nonnull-c-pointer nonnull-c-pointer))

(define uuid_unparse_lower
  (foreign-lambda void uuid_unparse_lower nonnull-c-pointer nonnull-c-pointer))

;;

(define-record-type uuid
  (%make-uuid variant puuid)
  %uuid?
  (variant uuid-variant set-uuid-variant!)
  (puuid %uuid-puuid))

(define-inline (unbox-puuid boxed-puuid)
  (%uuid-puuid boxed-puuid) )

(define-inline (boxed-puuid-compare uuid1 uuid2)
  (uuid_compare (unbox-puuid uuid1) (unbox-puuid uuid2)) )

(define (free-uuid uuid)
  (free (unbox-puuid uuid)) )

(define-inline (box-puuid puuid variant)
  (set-finalizer! (%make-uuid variant puuid) free-uuid) )

(define-inline (new-uuid variant)
  (box-puuid (allocate UUID-SIZEOF) variant) )

(define (puuid-version puuid)
  (bitwise-and (arithmetic-shift (pointer-u64-ref puuid) -52) #b1111) )

;;

(define (error-bad-argument-type loc obj #!optional msg)
  (##sys#signal-hook #:type-error loc
                     (apply string-append "bad argument type"
                                          (if msg (list " - " msg) '("")))
                     obj) )

(define (error-uuid-version loc obj)
  (error-bad-argument-type loc obj "invalid version") )

(define (error-uuid-variant loc obj)
  (error-bad-argument-type loc obj "invalid variant") )

(define (error-uuid-form loc obj)
  (error-bad-argument-type loc obj "not a string or symbol") )

(define (error-uuid-string-case loc obj)
  (error-bad-argument-type loc obj "invalid case") )

;;

(define (uuid-version->variant ver)
  (case ver
    ((0) #f)
    ((1) 'v1)
    ((2) 'v4)
    ((3) 'v3)
    ((4) 'v4)
    ((5) 'v5)
    (else (error-uuid-version 'uuid-version->variant ver))) )

(define (uuid-variant->version var)
  (case var
    ((#f) 0)
    (('v1) 1)
    (('v4) 2)
    (('v3) 3)
    (('v4) 4)
    (('v5) 5)
    (else (error-uuid-variant 'uuid-variant->version var))) )

(define (validate-uuid-variant loc obj)
  (case obj
    ((#f) #f)
    ((v1 V1 time) 'v1)
    ((v4 V4 random) 'v4)
    (else
      (error-uuid-variant loc obj))) )

(define (uuid? obj)
  (%uuid? obj) )

;optional @argnam is ignored
(define (error-uuid loc obj . _)
  (error-bad-argument-type loc obj "not a uuid") )

;optional @argnam is ignored
(define (check-uuid loc obj . _)
  (unless (%uuid? obj) (error-uuid loc obj))
  obj )

(define (uuid-v4-hash uuid bound randomization)
  ;"random" already
  (let ((puuid (unbox-puuid (check-uuid 'uuid-hash uuid))))
    (srfi69-hash-limit (pointer-u64-ref (pointer+ puuid 8)) bound randomization) ) )

(define (uuid-v1-hash uuid bound randomization)
  (let ((puuid (unbox-puuid (check-uuid 'uuid-hash uuid))))
    ;combine all bits (poorly)
    (srfi69-hash-limit (+ (pointer-u64-ref puuid)
                          (pointer-u64-ref (pointer+ puuid 8)))
                       bound randomization) ) )

(define (uuid-hash uuid #!optional (bound hash-default-bound)
                                   (randomization hash-default-randomization))
  ;(assert (= 16 UUID-SIZEOF))
  (case (uuid-variant uuid)
    ((#f)
      0 )
    ((v1)
      (uuid-v1-hash uuid bound randomization))
    ((v4)
      (uuid-v4-hash uuid bound randomization)) ) )

(define (uuid-null? obj)
  (uuid_is_null (unbox-puuid (check-uuid 'uuid-null? obj))) )

(define (*uuid-compare loc uuid1 uuid2)
  (boxed-puuid-compare (check-uuid loc uuid1) (check-uuid loc uuid2)) )

(define (uuid-compare uuid1 uuid2)
  (let ((cmp (*uuid-compare 'uuid-compare uuid1 uuid2)))
    (cond ((fxnegative? cmp) -1)
          ((fxzero? cmp) 0)
          (else 1) ) ) )

(define (uuid=? uuid1 uuid2)
  (fxzero? (*uuid-compare 'uuid=? uuid1 uuid2)) )

(define (uuid<? uuid1 uuid2)
  (fxnegative? (*uuid-compare 'uuid<? uuid1 uuid2)) )

(define (uuid>? uuid1 uuid2)
  (fxpositive? (*uuid-compare 'uuid>? uuid1 uuid2)) )

(define (uuid<=? uuid1 uuid2)
  (let ((cmp (*uuid-compare 'uuid<=? uuid1 uuid2)))
    (or (fxzero? cmp) (fxnegative? cmp))) )

(define (uuid>=? uuid1 uuid2)
  (let ((cmp (*uuid-compare 'uuid>=? uuid1 uuid2)))
    (or (fxzero? cmp) (fxpositive? cmp))) )

(define (uuid-clear! uuid)
  (uuid_clear (unbox-puuid (check-uuid 'uuid-clear! uuid)))
  (set-uuid-variant! uuid #f)
  uuid )

(define (uuid-copy old-uuid)
  (check-uuid 'uuid-copy old-uuid)
  (let ((uuid (new-uuid (uuid-variant old-uuid))))
    (uuid_copy (unbox-puuid uuid) (unbox-puuid old-uuid))
    uuid) )

(define (uuid-generate #!optional (variant #f))
  (let* ((variant (validate-uuid-variant 'uuid-generate variant))
         (uuid (new-uuid variant))
         (puuid (unbox-puuid uuid)) )
    (case variant
      ((#f)
        (uuid_generate puuid)
        (set-uuid-variant! uuid (uuid-version->variant (puuid-version puuid))))
      ((v1) (uuid_generate_time puuid))
      ((v4) (uuid_generate_random puuid)) )
    uuid ) )

(define (make-uuid . args)
  (if (null? args) (uuid-clear! (new-uuid #f))
      (apply uuid-generate args)) )

(define (uuid-parse uuid-text)
  (let* ((uuid-text (cond ((symbol? uuid-text) (symbol->string uuid-text))
                          ((string? uuid-text) uuid-text)
                          (else
                            (error-uuid-form 'uuid-parse uuid-text))))
         (uuid (new-uuid #f))
         (puuid (unbox-puuid uuid)) )
    (if (not (zero? (uuid_parse uuid-text puuid))) #f
        (begin
          (set-uuid-variant! uuid (uuid-version->variant (puuid-version puuid)))
          uuid) ) ) )

(define (uuid-unparse uuid #!optional (kase #f))
  (check-uuid 'uuid-unparse uuid)
  ;allocate string buffer, +1 for NUL byte; use string so gc
  (let* ((uuid-text (make-string (add1 UUID-STRING-LENGTH)))
         (puuid-text (make-locative uuid-text)) )
    (case kase
      ((#f)
        (uuid_unparse (unbox-puuid uuid) puuid-text))
      ((upper)
        (uuid_unparse_upper (unbox-puuid uuid) puuid-text))
      ((lower)
        (uuid_unparse_lower (unbox-puuid uuid) puuid-text))
      (else
        (error-uuid-string-case 'uuid-generate kase)) )
    ;remove NUL byte
    (substring uuid-text 0 UUID-STRING-LENGTH)) )

(define (string->uuid str)
  (uuid-parse str) )

(define (uuid->string uuid)
  (uuid-unparse uuid 'lower) )

(define (uuid-print obj #!optional (port (current-output-port)))
  (format port "#<~A ~A ~A>" uuid (uuid-variant obj) (uuid-unparse obj)) )

(set-record-printer! uuid uuid-print)

) ;module uuid-lib
